#!/usr/bin/ruby

# An irc bot able to find a shortest path from a word to another on wikipedia (fr)
# Author: Yoann Guillot, 03/2007
# licence: GPLv2

require 'libhttpclient'
require 'socket'
require 'sqlite3'
require 'cgi'

class WpLink

	def rq(url)
		@h.clear
		@h.get url, 0
	end

	def filt url
		url[/\/wiki\/(.*)/, 1]
	end

	# it does not working !!!
	def test_spam(title)
		rq '/wiki/' + title

		ina = false
		inlen, outlen = 0, 0
		acount = 0
		@h.curpage.parse.each { |e|
			ina = true if e.type == 'a'
			acount += 1 if e.type == 'a'
			ina = false if e.type == '/a'
			if e.type == 'String'
				if ina
					inlen += e.attr['content'].length
				else
					outlen += e.attr['content'].length
				end
			end

		}
#		puts "#{filt @h.cur_url}: acount: #{acount}, in #{inlen}, out #{outlen}, pc: " + "%.2f" % (inlen * 100.0 / (inlen + outlen))
		if acount >= 150 and inlen > outlen
			puts "spam detected for #{title}"
			true 
		end
	end

	def list_links
		@h.get_url_allowed.map { |u| filt u }.uniq.reject { |u| not u or u.empty? or u.include? ':' or
			u == 'Accueil' or u =~ /^\d*$/ or u == 'Homonymie' or u == 'Wiktionary' or 
			u =~ /^(\d+_)?(janvier|f.*vrier|mars|avril|mai|juin|juillet|ao.*t|septembre|octobre|novembre|d.*cembre)(_\d+)?$/i or
			u =~ /^Sigles_de_trois_lettres_de_/i
		}
	end

	def links_from(id)
		#puts "get links in "+find_name(id)
		dst_ids = @db.execute("select to_id from links where from_id=?", id).map{ |k| k[0] }
		#rq '/wiki/'+title
		#list_links
		return dst_ids
	end

	def links_to(id)
		#puts "get links to "+find_name(id)
		from_ids = @db.execute("select from_id from links where to_id=?", id).map { |k| k[0] }
		#rq '/w/index.php?title=Special:Whatlinkshere/'+title+'&limit=500&from=0'
		#list_links
		return from_ids
	end

	def walk(&b)
		loop do
			fwlen = @graph_src.length
			bklen = @graph_dst.length
			walk_fw(&b)
			walk_bk(&b)
			if fwlen == @graph_src.length or bklen == @graph_dst.length
				say "#{@graph_src.length + @graph_dst.length}: no way !"
				break
			end
			break if @stopping
			say "#{@graph_src.length + @graph_dst.length}..."
		end
	end

	def walk_fw
		todo = Hash.new()
		@graph_src.each { |k, v|
			if not v
				count = @db.get_first_value("select count(to_id) from links where from_id=?", k)
				todo[k] = count.to_i
			end
		}
		todo = todo.sort {|a,b| b[1]<=>a[1]}
		todo.each { |u,v|
#			next if test_spam(u)
			lf = @graph_src[u] = links_from(u)
			lf.delete_if { |l| @graph_src.has_key? l }
			lf.each { |l|
				throw :found, l if @graph_dst.has_key? l
				@graph_src[l] = nil
			}
			break if @stopping
			yield if block_given?
		}
	end

	def walk_bk
		todo = []
		@graph_dst.each { |k, v|
			todo << k if not v
		}
		todo.each { |u|
#			next if test_spam(u)
			lt = @graph_dst[u] = links_to(u)
			lt.delete_if { |l| @graph_dst.has_key? l }
			lt.each { |l|
				throw :found, l if @graph_src.has_key? l
				@graph_dst[l] = nil
			}
			break if @stopping
			yield if block_given?
		}
	end

	def search n
		case rq('/wiki/'+n).content
		when /Wikip.*dia ne poss.*de pas .*d.*article avec ce nom/, /Actuellement Wikip.*dia ne dispose d'aucun article concernant/
			puts "inexistant, search"
			rq '/wiki/Special:Search/'+n
			if not l = list_links.first or @h.curpage.content[/Pertinence : ([\d.]+)%/, 1].to_f < 10
				say "#{n} n'existe pas !"
				return
			end
		else
			l = filt @h.cur_url
		end
		l
	end

	def initialize(db)
		@host = 'fr.wikipedia.org'

		@h = HttpClient.new('fr.wikipedia.org')
		
		@db = SQLite3::Database.new(db)
		return if not @db
		@name_cache = Hash.new()
	end

	def find_name(id)
		res = @name_cache[id]
		return res if res
		res = @db.get_first_value("select title from page where page_id=?", id)
		if not res then
			res = ""
			puts "!!!!!! "+id
		end
		@name_cache[id] = res
		return res
	end

	def fnd_rec(h, s, d)
		if s == d
			[find_name(d)]
		elsif h[s] and h[s].to_a.include? d
			[find_name(s), find_name(d)]
		else
			l = nil
			h[s].find { |u|
				h[u] and l = fnd_rec(h, u, d)
			}
			[find_name(s)] + l if l
		end
	end

	attr_reader :solving
	def stop
		@stopping = true
	end

	def solve(src, dst, &b)
		@solving = true
		@stopping = false
		begin
			src = search src.delete(':/')
			dst = search dst.delete(':/')
			return if not src or not dst
			say "searching #{src} -> #{dst}..."
			src_id = @db.get_first_value("select page_id from page where title=?", CGI::unescape(src))
			dst_id = @db.get_first_value("select page_id from page where title=?", CGI::unescape(dst))

			if not src_id or not dst_id then
				say "not found !"
				return
			end
				
			@graph_src = {}
			@graph_dst = {}
			@graph_src[src_id] = nil
			@graph_dst[dst_id] = nil
			
			if match = catch(:found) { walk(&b) }
				path = fnd_rec(@graph_src, src_id, match) + fnd_rec(@graph_dst, dst_id, match).reverse[1..-1]
				say "found (#{@graph_src.length + @graph_dst.length}): " + path.join(' -> ')
			end
		ensure
			@solving = false
		end
	end
end

class IrcBot
	attr_reader :chan
	def initialize(server, chan, nick, db)
		@server, @chan, @nick = server, chan, nick
		@s = TCPSocket.open(@server, 6667)
		@wp = WpLink.new db
	end

	def send a
		@s.write a + "\r\n"
	end

	def pm a
		send "PRIVMSG #@chan :#{a}"
	end

	def main_loop
		connect
		loop { do_things }
	end

	def connect
		send "user wpbot wpbot wpbot wpbot"
		send "nick "+@nick
		send "join #@chan"
	end

	def do_things
		case l = @s.gets.chomp
		when /^PING (.*)/
			send "PONG #$1"
		when /PRIVMSG #@chan :!wp\s+(.*)/
			puts l
			case $1
			when /(\S+)\s+(\S+)/
				if @wp.solving
					pm "i'm busy !"
				else
					solve($1, $2)
				end
			when /stop/
				@wp.stop
				pm "ok"
			when /help/
				pm "usage: !wp help | !wp stop | !wp <src> <dst>"
			when /reload/
				load $0
			end
		when / (PRIVMSG|MODE|KICK|JOIN|PART|QUIT|NICK) /
		else
			puts l
		end
	end

	def solve src, dst
		@wp.solve(src, dst) { do_things while IO.select([@s], nil, nil, 0) }
	end
end

# booh
def say a
	$bot.send "PRIVMSG #{$bot.chan} :#{a.gsub(/%(..)/) { $1.hex.chr }}"
	puts a
end

if not defined? $started
	        $started = true

		$bot = IrcBot.new ARGV.shift, ARGV.shift, ARGV.shift, ARGV.shift
		$bot.main_loop
end
