#!/bin/bash

pagelinks="frwiki-latest-pagelinks.sql.gz"
page="frwiki-latest-page.sql.gz"

database="wp2.db"

for filename in $pagelinks $page ; do
	if [ ! -f "$filename" ]; then
		echo "File $filename is missing, exiting ..."
		exit 1
	fi
done

[ -f "$database" ] && rm "$database"
[ -f "debug.log" ] && rm "debug.log"

(
	cat <<EOF
PRAGMA auto_vacuum = 1;
PRAGMA cache_size = 40000;
PRAGMA synchronous = OFF;

CREATE TEMPORARY TABLE pagelinks
(
	from_id INTEGER,
	title TEXT
);

CREATE TABLE links
(
	from_id INTEGER,
	to_id INTEGER
);

CREATE TABLE page
(
	page_id INTEGER,
	title TEXT
);

BEGIN;
EOF

	# Converting pagelinks
	zcat "$pagelinks" |
	grep INSERT |
	./a.out pagelinks |
	# Keeping only links from pages in namespace 0
	egrep 'VALUES \([0-9]+,0' |
	sed 's/VALUES (\([0-9]\+\),0,/VALUES (\1,/'

	# Converting pages
	zcat "$page" |
	grep INSERT |
	./a.out page |
	# Keeping only links from pages in namespace 0
	egrep 'VALUES \([0-9]+,0' |
	sed 's/VALUES (\([0-9]\+\),0,\(.*\),.*,.*,.*,.*,.*,.*,.*,.*,.*);$/VALUES (\1,\2);/'

	cat <<EOF
COMMIT;

CREATE INDEX page_page_id ON page (page_id);
CREATE INDEX page_title ON page (title);

INSERT INTO links SELECT l.from_id AS from_id, p.page_id AS to_id FROM page p, pagelinks l WHERE (p.title = l.title);

CREATE INDEX links_from_id ON links (from_id);
CREATE INDEX links_to_id ON links (to_id);

EOF
) | sqlite3 "$database"
